############################################################################################
# app.R
#
# Examples of the usage of the algorithm using synthetic data.
#
# Code by
#
#  - Ricardo Silva (ricardo@stats.ucl.ac.uk)
#
# Current version: 03/01/2017
# First version: 03/01/2017

# Example of usage:
#
# First, generate and run synthetic studies for discovering instumental variables:
#
# > iv_discover_example(U_var_1 = 0.125, U_var_2 = 0.125, N = 10000, N_max = 10000)
#
# Then, verify summary of output:
#
# > iv_discover_example_eval(num_best = 15, U_var1 = 0.125, U_var2 = 0.125, N_max = 10000)
#
# To reproduce Table 1 in the paper "Learning Instrumental Variables with Structural 
# and Non-Gaussianity Assumptions", change N_max to 5000 and 1000. Also, change
# U_var_1 and U_var_2 to 0.25.
# 
# The following generates synthetic studies for assessing the non-Gaussianity tests
# in pre-defined situations:
#
# > iv_lingamtest_example(U_var = 0.125, N = 10000, N_max = 10000)
#
# Then, verify summary of output:
#
# > iv_lingamtest_example_eval(U_var = 0.125, N_max = 10000)

#' @title Example of usage of the \code{IV-TETRAD} algorithm for causal effect estimation.
#' 
#' @description
#' 
#' Generates synthetic examples and evaluation of different algorithms for estimating
#' causal effects in linear models using instrumental variables.
#' 
#' @param setups an object of class \code{\link{batch_benchmark_problems}}.
#'
#' @param U_var_1         proportion of variance in treatment and outcome due to their latent common causes possible connected to colliders
#' @param U_var_2         proportion of variance in treatment and outcome due to their latent common causes not connected to colliders.
#' @param C               number of elements in each spurious cluster
#' @param K               number of spurious clusters
#' @param num_direct      number of elements in cluster that are parents of Y
#' @param num_possible    number of elements in possible solution, but only if we do not condition on colliders
#' @param num_best        number of elements in "true cluster"
#' @param num_colliders   number of colliders destroying the tetrads of the "possible set"
#' @param N               sample size
#' @param K_win           number of variables to seed the swipe search
#' @param num_simulations number of simulations
#' @param N_max           which subset of generate points to actually use in the experiments 
#' @param save_data       whether to save problem instances into files 
#' @param file_path       directory where to store the file.
#'
#' @return This does not return any value, instead it saves results to a file.
#'
#' @details
#' This provides a full example on how to run a simulation study using the main method
#' developed in this package, along with a comparison against standard procedures. See
#' the description of \code{run_experiment} for more details.
#'  
#' @export

iv_discover_example <- function(
  U_var_1 = 0.125,  U_var_2 = 0.125, 
  C = 10,  K = 1, num_direct = 5, num_possible = 10, num_best = 15, num_colliders = 10,
  N = 10000, K_win = 5, num_simulations = 100, N_max = 10000, save_data = FALSE, file_path = "./")
{
  # Load seed
  
  load("seed_file")
  set.seed(seed)
  
  # Generate problem files 
  
  problems <- list()
  for (i in seq_len(num_simulations)) {
    problems[[i]] <- simulate_partition_problem(C, K, num_direct, num_possible, num_colliders, num_best, U_var_1, U_var_2)
    problems[[i]]$data <- generate_data(N, problems[[i]], laplace_error = TRUE)
  }
  if (save_data) {
    # (WARNING: THIS MAY TAKE MUCH DISK SPACE)
    filename <- paste(file_path, "problems_G", num_best, "_N", N, "_U1_", U_var_1 * 100, "_U2_", U_var_2 * 100, ".Rdata", sep = "")
    save(problems, file = filename)
  }
  
  # Run search algorithms, save results
  
  out <- run_experiment(problems, K_win, N_max)
  filename <- paste(file_path, "results_G", num_best, "_N", N_max, "_U1_", U_var_1 * 100, "_U2_", U_var_2 * 100, ".Rdata", sep = "")
  save(out, file = filename)  
  
}

#' @title Run a series of synthetic experiments, and evaluate them.
#' 
#' @description
#' 
#' This takes as inputs a series of problems generated with function \code{\link{iv_discover_example}}, and report the
#' evaluation of several methods. The goal is to estimate average causal effects (ACEs) for each problem.
#' 
#' @param problems an object generated by function \code{iv_discover_example}.
#' @param K        number of variables to keep in the IV-TETRAD search.
#' @param N_max    if provided, crops the dataset at N_max rows.
#' @param verbose  if \code{TRUE}, prints detailed progression of the experiment.
#'
#' @return A list containing several outcomes of interest:
#'   \item{\code{true_ACEs}}{an array containing the true ACEs for each problem.}
#'   \item{\code{naive_ACEs1}}{estimated ACEs given by the naive regression of outcome on treatment.}
#'   \item{\code{naive_ACEs2}}{estimated ACEs given by using all covariates as instrumental variables.}
#'   \item{\code{naive_ACEs3}}{estimated ACEs given by back-door adjustment on all covariates.}
#'   \item{\code{oracle_ACEs}}{estimated ACEs using the correct set of instrumental variables and adjustment set.}
#'   \item{\code{sisvive_ACEs}}{estimated ACEs using \code{sisVIVE}.}
#'   \item{\code{S_oracle_ACEs}}{estimated ACEs using \code{sisVIVE} excluding all covariates known to add bias.}
#'   \item{\code{W_oracle_ACEs}}{estimated ACEs using the correct set of instrumental variables and all remaining variables as adjustment set.}
#'   \item{\code{sisvive_selection}}{instruments selected by \code{sisVIVE}.}
#'   \item{\code{tetrad_iv_ACEs}}{estimated ACEs using \code{IV-TETRAD}.}
#'   \item{\code{tetrad_iv_output}}{objects generated by the \code{IV-TETRAD} search.}
#'   \item{\code{K_win}}{copy of the \code{K_win} parameter given as input.}
#'
#' @details
#' A variety of methods are executed, with the corresponding estimated ACEs report as a list of arrays.
#' 
#' @export

run_experiment <- function(problems, K_win, N_max = NULL, verbose = TRUE)
{  
  num_simulations <- length(problems)
  
  true_ACEs       <- rep(0, num_simulations)  # True ACEs
  naive_ACEs1     <- rep(0, num_simulations)  # Regress Y on X directly
  naive_ACEs2     <- rep(0, num_simulations)  # Regress on (W and Z) and empty set  
  naive_ACEs3     <- rep(0, num_simulations)  # Regress on empty set and (W and Z)
  oracle_ACEs     <- rep(0, num_simulations)  # TSLS (two-stage least squares) with correct W and Z
  sisvive_ACEs    <- rep(0, num_simulations)  # sisVIVE, starting with W and all of Z
  S_oracle_ACEs   <- rep(0, num_simulations)  # sisVIVE, starting with W and valid Z only
  W_oracle_ACEs   <- rep(0, num_simulations)  # Regress on true W and all other Z
  tetrad_iv_ACEs  <- rep(0, num_simulations)  # IV-TETRAD
  
  sisvive_selection <- list()
  tetrad_iv_output  <- list()
  tetrad_tests      <- list()
  
  X_idx     <- problems[[1]]$X_idx
  Y_idx     <- problems[[1]]$Y_idx
  g         <- problems[[1]]$G
  W_space   <- c(problems[[1]]$W_spurious_idx, problems[[1]]$W_direct_idx, problems[[1]]$W_possible_idx, problems[[1]]$W_colliders_idx, problems[[1]]$W_best_idx)
  t_W_space <- c(problems[[1]]$W_spurious_idx, problems[[1]]$W_direct_idx, problems[[1]]$W_possible_idx, problems[[1]]$W_best_idx)
  true_W    <- c(problems[[1]]$W_possible_idx, problems[[1]]$W_best_idx)
  true_Z    <- c(problems[[1]]$W_spurious_idx, problems[[1]]$W_direct_idx)
  all_other <- c(true_Z, problems[[1]]$W_colliders_idx)
  
  for (i in 1:num_simulations) {
    
    if (verbose) cat("Simulation", i, "\n")
    
    dat <- problems[[i]]$data
    if (!is.null(N_max)) {
      dat <- dat[1:N_max, ]
    }
    m <- matrix(colMeans(dat), nrow = 1)
    dat <- dat - m[rep(1, nrow(dat)), ]
    cov_dat <- cov(dat)
    
    true_ACEs[i] <- problems[[i]]$model$L[Y_idx, X_idx]
    if (verbose) cat(sprintf("  True ACE       = %+.2f\n", true_ACEs[i]))
    
    naive_ACEs1[i] <- cov_dat[Y_idx, X_idx] / cov_dat[X_idx, X_idx]
    if (verbose) cat(sprintf("  Naive1 ACE     = %+.2f (%+.2f)\n", naive_ACEs1[i], naive_ACEs1[i] - true_ACEs[i]))
    
    naive_ACEs2[i] <- IV_TSLS(W_space, c(), X_idx, Y_idx, cov_dat)
    if (verbose) cat(sprintf("  Naive2 ACE     = %+.2f (%+.2f)\n", naive_ACEs2[i], naive_ACEs2[i] - true_ACEs[i]))
    
    other_X <- c(X_idx, W_space)    
    naive_ACEs3[i] <- solve(cov_dat[other_X, other_X], cov_dat[other_X, Y_idx])[1]
    if (verbose) cat(sprintf("  Naive3 ACE     = %+.2f (%+.2f)\n", naive_ACEs3[i], naive_ACEs3[i] - true_ACEs[i]))
    
    oracle_ACEs[i] <- IV_TSLS(true_W, true_Z, X_idx, Y_idx, cov_dat)
    if (verbose) cat(sprintf("  Oracle ACE     = %+.2f (%+.2f)\n", oracle_ACEs[i], oracle_ACEs[i] - true_ACEs[i]))
    
    W_oracle_ACEs[i] <- IV_TSLS(true_W, all_other, X_idx, Y_idx, cov_dat)
    if (verbose) cat(sprintf("  W-Oracle ACE   = %+.2f (%+.2f)\n", W_oracle_ACEs[i], W_oracle_ACEs[i] - true_ACEs[i]))
    
    iv_s_oracle <- cv.sisVIVE(Y = dat[, Y_idx], D = dat[, X_idx], Z = dat[, t_W_space])
    S_oracle_ACEs[i] <- iv_s_oracle$beta
    if (verbose) cat(sprintf("  S-Oracle ACE   = %+.2f (%+.2f)\n", S_oracle_ACEs[i], S_oracle_ACEs[i] - true_ACEs[i]))
    
    iv_sisvive <- cv.sisVIVE(Y = dat[, Y_idx], D = dat[, X_idx], Z = dat[, W_space])
    sisvive_ACEs[i] <- iv_sisvive$beta
    sisvive_selection[[i]] <- W_space[which(iv_sisvive$alpha == 0)]
    if (verbose) cat(sprintf("  sisVIVE ACE    = %+.2f (%+.2f)\n", sisvive_ACEs[i], sisvive_ACEs[i] - true_ACEs[i]))
    
    tetrad_hat <- iv_tetrad(dat, W_space, X_idx, Y_idx, K_win)
    tetrad_iv_output[[i]] <- tetrad_hat
    tetrad_iv_ACEs[i] <- tetrad_hat$ace
    if (verbose) cat(sprintf("  TETRAD-IV ACE  = %+.2f (%+.2f)\n", tetrad_iv_ACEs[i], tetrad_iv_ACEs[i] - true_ACEs[i]))
    
  }
  
  out <- list(true_ACEs = true_ACEs, naive_ACEs1 = naive_ACEs1, naive_ACEs2 = naive_ACEs2, naive_ACEs3 = naive_ACEs3, 
              oracle_ACEs = oracle_ACEs, sisvive_ACEs = sisvive_ACEs, S_oracle_ACEs = S_oracle_ACEs,  W_oracle_ACEs = W_oracle_ACEs, 
              sisvive_selection = sisvive_selection, tetrad_iv_ACEs = tetrad_iv_ACEs, tetrad_iv_output = tetrad_iv_output, 
              K_win = K_win)
  return(out)
}

#' @title Summarizes output logs generated by simulated study of causal effect estimation
#' by instrument detection.
#' 
#' @description
#' 
#' Reads the output files generated by a call to \code{iv_discover_example} and
#' provides a comparative summary of several algorithms.
#' 
#' @param num_best  number of elements in "true cluster" in the simulated study.
#' @param U_var_1   proportion of variance in treatment and outcome due to their latent common causes possible connected to colliders.
#' @param U_var_2   proportion of variance in treatment and outcome due to their latent common causes not connected to colliders.
#' @param N_max     which subset of generate points to actually use in the experiments.
#' @param file_path directory where to find the file.
#'
#' @details
#' This provides a simple summary of the output of \code{run_experiment}. See the documentation
#' of that function for more details.
#'  
#' @export

iv_discover_example_eval <- function(num_best, U_var_1, U_var_2, N_max, file_path = "./")
{
  filename <- paste(file_path, "results_G", num_best, "_N", N_max, "_U1_", U_var_1 * 100, "_U2_", U_var_2 * 100, ".Rdata", sep = "")
  load(file = filename) 
  compare_pvalue <- binom.test(sum(abs(out$true_ACEs - out$tetrad_iv_ACEs) < abs(out$true_ACEs - out$naive_ACEs2)), length(out$true_ACEs))$p.value
  cat("* ABSOLUTE ERROR:\n")
  cat(sprintf("Naive 1  : %.2f\n", median(abs(out$true_ACEs - out$naive_ACEs1))))
  cat(sprintf("Naive 2  : %.2f (p value = %.1f)\n", median(abs(out$true_ACEs - out$naive_ACEs2)), compare_pvalue <- binom.test(sum(abs(out$true_ACEs - out$tetrad_iv_ACEs) < abs(out$true_ACEs - out$naive_ACEs2)), length(out$true_ACEs))$p.value))
  cat(sprintf("Naive 3  : %.2f\n", median(abs(out$true_ACEs - out$naive_ACEs3))))
  cat(sprintf("Oracle   : %.2f\n", median(abs(out$true_ACEs - out$oracle_ACEs))))
  cat(sprintf("S-Oracle : %.2f\n", median(abs(out$true_ACEs - out$S_oracle_ACEs))))
  cat(sprintf("W-Oracle : %.2f\n", median(abs(out$true_ACEs - out$W_oracle_ACEs))))
  cat(sprintf("SISVIVE  : %.2f\n", median(abs(out$true_ACEs - out$sisvive_ACEs))))
  cat(sprintf("IV-TETRAD: %.2f\n", median(abs(out$true_ACEs - out$tetrad_iv_ACEs))))
  cat("\n")
  cat("* SIGN PREDICTION:\n")
  cat(sprintf("Naive 1  : %.2f\n", mean(out$true_ACEs * out$naive_ACEs1 > 0)))
  cat(sprintf("Naive 2  : %.2f\n", mean(out$true_ACEs * out$naive_ACEs2 > 0)))
  cat(sprintf("Naive 3  : %.2f\n", mean(out$true_ACEs * out$naive_ACEs3 > 0)))
  cat(sprintf("Oracle   : %.2f\n", mean(out$true_ACEs * out$oracle_ACEs > 0)))
  cat(sprintf("S-Oracle : %.2f\n", mean(out$true_ACEs * out$S_oracle_ACEs > 0)))
  cat(sprintf("W-Oracle : %.2f\n", mean(out$true_ACEs * out$W_oracle_ACEs > 0)))
  cat(sprintf("SISVIVE  : %.2f\n", mean(out$true_ACEs * out$sisvive_ACEs > 0)))
  cat(sprintf("IV-TETRAD: %.2f\n", mean(out$true_ACEs * out$tetrad_iv_ACEs > 0)))
}

#' @title Example of usage of the non-Gaussianity test for instrumental variable validation.
#' 
#' @description
#' 
#' This function generates synthetic studies for three different scenarions.
#' 
#' @param setups an object of class \code{\link{batch_benchmark_problems}}.
#'
#' @param U_var           proportion of variance in treatment and outcome due to their latent common causes possible connected to colliders.
#' @param N               sample size.
#' @param p_value         p-value for the tests.
#' @param num_simulations number of simulations.
#' @param N_max           which subset of generate points to actually use in the experiments.
#' @param save_data       whether to save problem instances into files. 
#' @param file_path       directory where to find the file.
#'
#' @return This does not return any value, instead it saves results to a file.
#'
#' @details
#' This provides a full example on how to run a simulation study using the main method
#' developed in this package, along with a comparison against standard procedures. See
#' the description of \code{run_experiment} for more details. There are three types of
#' problems.
#' 
#' Problems of Type 1 have the following structure: three latent variables U_0, U_XY1, U_XY2;
#' three observed auxiliary variables W1, W2 and Z. The graph is given by
#'     W1 <- U_0
#'     W2 <- U_0
#'     Z  <- W1, W2, U_XY1
#'     X  <- U_XY1, U_XY2
#'     Y  <- X, U_XY1, U_XY2
#' The ideal output is to falsify the null hypothesis of residual independence between
#' W_i and X given {W_j, Z}. This will be easier the more non-Gaussian the data is
#' (hence the call to generate data with the flag 'laplace_error = FALSE'.) It will also
#' be easier if U_XY1 carries most of the unmeasured confounding variability. In the script
#' below, U_XY2 is basically ignored.
#' 
#' Problems of Type 2 have the following structure: three latent variables U_0, U_XY1, U_XY2;
#' two observed auxiliary variables W1, W2. The graph is given by
#'     W1    <- U_0
#'     W2    <- U_0
#'     U_XY1 <- U_0
#'     X     <- U_XY1, U_XY2
#'     Y     <- X, U_XY1, U_XY2
#' The ideal output is to falsify the null hypothesis of residual independence between
#' W_i and X given W_j. This will be easier the more non-Gaussian the data is
#' (hence the call to generate data with the flag 'laplace_error = FALSE'.) It will also
#' be easier if U_XY1 carries most of the unmeasured confounding variability. In the script
#' below, U_XY2 is basically ignored.
#'
#' Problems of Type 3 have the following structure: three latent variables U_0, U_XY1, U_XY2;
#' two observed auxiliary variables W1, W2. The graph is given by
#'     W1    <- U_0
#'     W2    <- U_0
#'     X     <- W1, W2, U_XY1, U_XY2
#'     Y     <- X, U_XY1, U_XY2
#' The ideal output is NOT to falsify the null hypothesis of residual independence between
#' W_i and X given W_j. This will be easier the more non-Gaussian the data is
#' (hence the call to generate data with the flag 'laplace_error = FALSE'.) It will also
#' be easier if U_XY1 carries most of the unmeasured confounding variability. In the script
#' below, U_XY2 is basically ignored.
#'
#' @export

iv_lingamtest_example <- function(U_var = 0.125, N = 10000, p_value = 0.05,       
                         num_simulations = 100, N_max = 10000, save_data = FALSE, file_path = "./")
{
  # Load seed
  
  load("seed_file")
  set.seed(seed)
  
  # Generate problem files 
  
  U_var_1 <- U_var          # proportion of variance in treatment and outcome due to their latent common causes possible connected to colliders.
  U_var_2 <- 0.01           # Proportion of variance in treatment and outcome due to their latent common causes not connected to colliders.
  C <- 0                    # Number of elements in each spurious cluster
  K <- 1                    # Number of spurious clusters
  num_direct <- 0           # Number of elements in cluster that are parents of Y
  num_possible <- 2         # Number of elements in possible solution, but only if we do not condition on colliders
  num_best <- 0             # Number of elements in "true cluster"
  num_colliders <- 1        # Number of colliders destroying the tetrads of the "possible set"
  
  problems1 <- list()
  for (i in seq_len(num_simulations)) {
    problems1[[i]] <- simulate_partition_problem(C, K, num_direct, num_possible, num_colliders, num_best, U_var_1, U_var_2, direct_X = FALSE)
    problems1[[i]]$data <- generate_data(N, problems1[[i]], FALSE)
  }
  problems2 <- list()
  for (i in seq_len(num_simulations)) {
    problems2[[i]] <- simulate_confounded_problem_unblockable(num_possible, U_var_1, U_var_2)
    problems2[[i]]$data <- generate_data(N, problems2[[i]], FALSE)
  }
  problems3 <- list()
  for (i in seq_len(num_simulations)) {
    problems3[[i]] <- simulate_confounded_problem_blockable(num_possible, U_var_1, U_var_2)
    problems3[[i]]$data <- generate_data(N, problems3[[i]], FALSE)
  }
  if (save_data) {
    # (THIS MAY TAKE MUCH DISK SPACE)
    filename <- paste(file_path, "problems_test_N", N, "_U_", U_var_1 * 100, ".Rdata", sep = "")
    save(problems1, problems2, problems3, file = filename)
  }
  
  # Run search algorithms, save results

  cat("BATCH 1\n"); out1 <- run_experiment_detection(problems1, p_value, N_max)  
  cat("BATCH 2\n"); out2 <- run_experiment_detection(problems2, p_value, N_max)
  cat("BATCH 3\n"); out3 <- run_experiment_detection(problems3, p_value, N_max)
  filename <- paste(file_path, "results_test_N", N_max, "_U_", U_var_1 * 100, ".Rdata", sep = "")
  save(out1, out2, out3, file = filename)  
  
}

#' @title Run a series of synthetic experiments for evaluating non-Gaussianity tests.
#' 
#' @description
#' 
#' This takes as inputs a series of problems generated with function \code{\link{iv_lingamtest_example}}, and report the
#' evaluation of several methods. The goal is to estimate average causal effects (ACEs) for each problem.
#' 
#' @param problems an object generated by function \code{iv_lingamtest_example}.
#' @param p_value  p-value for the tests.
#' @param N_max    which subset of generate points to actually use in the experiments.
#' @param verbose  if \code{TRUE}, prints detailed progression of the experiment.
#'
#' @return A list containing several outcomes of interest:
#'   \item{\code{true_ACEs}}{an array containing the true ACEs for each problem.}#'   
#'   \item{\code{tetrad_test_output}}{decisions returned by the non-Gaussianity tests in each problem.}
#'   \item{\code{p_value}}{the p-value threshold given as input to the function.}
#'
#' @details
#' Decisions are performed separately for each instrument candidate, and an overall decision returned.
#' 
#' @export

run_experiment_detection <- function(problems, p_value, N_max = NULL, verbose = TRUE)
{  
  num_simulations <- length(problems)
  
  true_ACEs       <- rep(0, num_simulations)  # True ACEs
  
  X_idx     <- problems[[1]]$X_idx
  Y_idx     <- problems[[1]]$Y_idx
  
  W_candidate <- problems[[1]]$W_possible_idx
  Z_candidate <- problems[[1]]$W_colliders_idx
  tetrad_test_output <- list()
  
  for (i in 1:num_simulations) {
    
    if (verbose) cat("Simulation", i, "\n")
    
    dat <- problems[[i]]$data    
    if (!is.null(N_max)) {
      dat <- dat[1:N_max, ]
    }
    dat <- standardize_data(dat)
    
    true_ACEs[i] <- problems[[i]]$model$L[Y_idx, X_idx]
    if (verbose) cat(sprintf("  True ACE       = %+.2f\n", true_ACEs[i]))
    
    tetrad_test_output[[i]] <- lingam_iv_test(dat, W_candidate, Z_candidate, X_idx, Y_idx, p_value, verbose = TRUE)
    if (verbose) cat(sprintf("  LiNGAM-IV test  = %.2f\n", tetrad_test_output[[i]]$test_summary))
    
  }
  
  out <- list(true_ACEs = true_ACEs, tetrad_test_output = tetrad_test_output, p_value = p_value)
  return(out)
}

#' @title Summarizes output logs generated by simulated study of non-Gaussianity tests of
#' candidate instruments.
#' 
#' @description
#' 
#' Reads the output files generated by a call to \code{iv_lingamtest_example} and
#' provides a summary of the results based on the desired outcomes
#' 
#' @param U_va1     proportion of variance in treatment and outcome due to their latent common causes possible connected to colliders.
#' @param N_max     which subset of generate points to actually use in the experiments.
#' @param file_path directory where to find the file.
#'
#' @details
#' This provides a simple summary of the output of three variations \code{run_experiment_detection}, as
#' determined by function \code{iv_lingamtest_example}. See the documentation of those functions for more details.
#'  
#' @export

iv_lingamtest_example_eval <- function(U_var, N_max, file_path = "./")
{
  filename <- paste(file_path, "results_test_N", N_max, "_U_", U_var * 100, ".Rdata", sep = "")
  load(file = filename)
  correct <- rep(0, 3)
  num_trials <- length(out1$true_ACEs)
  for (i in seq_len(num_trials)) {
    result1 <- out1$tetrad_test_output[[i]]$test_summary
    if (result1 <= 0.5) correct[1] <- correct[1] + 1
    result2 <- out2$tetrad_test_output[[i]]$test_summary
    if (result2 <= 0.5) correct[2] <- correct[2] + 1
    result3 <- out3$tetrad_test_output[[i]]$test_summary
    if (result3 > 0.5) correct[3] <- correct[3] + 1
  }
  cat(sprintf("Results with sample size %d:\n", N_max))
  cat(sprintf("  - Proportion correct, Setup 1  : %.2f\n", correct[1] / num_trials))
  cat(sprintf("  - Proportion correct, Setup 2  : %.2f\n", correct[2] / num_trials))
  cat(sprintf("  - Proportion correct, Setup 3  : %.2f\n", correct[3] / num_trials))
}
