############################################################################################
# app_sachs.R
#
# Exemplifies the use of IV-TETRAD using data from
#
# "Causal Protein-Signaling Networks Derived from Multiparameter Single-Cell Data"
#   Karen Sachs, Omar Perez, Dana Pe'er, Douglas A. Lauffenburger, Garry P. Nolan
#   Science  22 Apr 2005:
#   Vol. 308, Issue 5721, pp. 523-529
#
# Data can be downloaded from http://science.sciencemag.org/content/308/5721/523.full.
#
# Code by
#
#  - Ricardo Silva (ricardo@stats.ucl.ac.uk)
#
# Current version: 03/01/2017
# First version: 03/01/2017

# Prepare environment

library(gdata) # To read Excel files. Unnecessary if properly transformed in csv first
dat <- read.xls("1. cd3cd28.xls", header = TRUE)
var_names <- names(dat)
n <- nrow(dat)
cov_dat <- cov(dat)
X_idx <- which(var_names == "p44.42")
Y_idx <- which(var_names == "pakts473")
W_scope <- setdiff(1:ncol(dat), c(X_idx, Y_idx))
num_ivs <- 3
dat_mat <- as.matrix(dat)

load("seed_file")
set.seed(seed)

# Non-Bayesian

iv_sisvive <- cv.sisVIVE(Y = dat_mat[, Y_idx], D = dat_mat[, X_idx], Z = dat_mat[, W_scope])
naive1 <- cov_dat[Y_idx, X_idx] / cov_dat[X_idx, X_idx]
naive2 <- IV_TSLS(W_scope, c(), X_idx, Y_idx, cov_dat)
naive3 <- solve(cov_dat[c(X_idx, W_scope), c(X_idx, W_scope)], cov_dat[c(X_idx, W_scope), Y_idx])[1]
tetrad_result <- iv_tetrad(dat_mat, W_scope, X_idx, Y_idx, num_ivs, blanket_prune = FALSE)

cat("  Naive1    =", naive1, "\n")
cat("  Naive2    =", naive2, "\n")
cat("  Naive3    =", naive3, "\n")
cat("  sisVIVE   =", iv_sisvive$beta, "\n")
cat("  IV-TETRAD =", tetrad_result$ace, "\n")

# Bootstrap

num_boot <- 500
ace_sisvive <- rep(0, num_boot)
ace_tetrad <- rep(0, num_boot)
ace_regression <- rep(0, num_boot)
for (i in seq_len(num_boot)) {
  cat(sprintf("[%d] Bootstrapping\n", i))
  bdat <- dat_mat[sample(n, n, replace = TRUE), ]
  ace_sisvive[i] <- cv.sisVIVE(Y = bdat[, Y_idx], D = bdat[, X_idx], Z = bdat[, W_scope])$beta
  ace_tetrad[i] <- iv_tetrad(bdat, W_scope, X_idx, Y_idx, num_ivs, blanket_prune = FALSE)$ace
  cov_bdat <- cov(bdat)
  ace_regression[i] <- solve(cov_bdat[c(X_idx, W_scope), c(X_idx, W_scope)], cov_bdat[c(X_idx, W_scope), Y_idx])[1]
}
xl <- c(min(ace_tetrad), max(ace_tetrad))
x11(); plot(density(ace_sisvive), xlim = xl, xlab = "differential causal effect", main = "sisVIVE bootstrapped empirical distribution")
x11(); plot(density(ace_tetrad), xlim = xl, xlab = "differential causal effect", main = "IV-TETRAD bootstrapped empirical distribution")
x11(); plot(density(ace_regression), xlim = xl, xlab = "differential causal effect", main = "Regression bootstrapped empirical distribution")

#x11(); plot(density(ace_sisvive), xlim = xl, xlab = "differential causal effect", main = "sisVIVE", cex.main = 3.5, cex.lab = 2, cex.axis = 2)
#x11(); plot(density(ace_tetrad), xlim = xl, xlab = "differential causal effect", main = "IV-TETRAD", cex.main = 3.5, cex.lab = 2, cex.axis = 2)
#x11(); plot(density(ace_regression), xlim = xl, xlab = "differential causal effect", main = "NAIVE3", cex.main = 3.5, cex.lab = 2, cex.axis = 2)

# LiNGAM tests

lingam_tests <- list()
for (i in seq_along(tetrad_result$W)) {
  cat(sprintf("LiNGAM tests for solution %d,\n", i))
  lingam_tests[[i]] <- lingam_iv_test(dat_mat, tetrad_result$W[[i]], tetrad_result$Z[[i]], X_idx, Y_idx, 0.05, verbose = FALSE)
  for (j in seq_along(lingam_tests[[i]]$residual_tests)) {
    cat(sprintf("  instrument %s: pvalue = %.2f\n", var_names[tetrad_result$W[[i]][j]], lingam_tests[[i]]$residual_tests[j]))  
  }
}

# Bayesian: initial parameters

num_clusters <- 5    # Number of mixture components
dirichlet_ess <- 10  # Equivalent sample size used by Dirichlet prior over distribution of mixture components
base_model <- list(ncomp = num_clusters, A = matrix(1, ncol = 1), a = rep(dirichlet_ess / num_clusters, num_clusters))

# Investigate sampling from the prior

mcmc_params_prior <- list(num_steps = 1000, burn_in = 0, thinning = 1)
alpha_cut <- 10
tetrad_threshold <- 0.01
out_prior <- iv_tetrad_bayes(matrix(nrow = 0, ncol = length(var_names)), X_idx, Y_idx, num_ivs, alpha_cut, tetrad_threshold, base_model, mcmc_params_prior, verbose = 1)
x11(); plot(density(out_prior$ace), xlab = "differential causal effect", ylab = "Prior density", main = paste("IV-TETRAD (", alpha_cut, ",", tetrad_threshold, ")", sep = ""), cex.main = 3.5, cex.lab = 2, cex.axis = 2, xlim = c(-5, 5))

# Do inference with the data

mcmc_params <- list(num_steps = 10500, burn_in = 500, thinning = 5)
out <- iv_tetrad_bayes(dat_mat, X_idx, Y_idx, num_ivs, alpha_cut, tetrad_threshold, base_model, mcmc_params_prior, verbose = 1, store_model = TRUE)

# Summarize uncertainty

IV_freq <- bayes_iv_summary(out)
IV_bayes_score <- colMeans(IV_freq)
IV_ranked <- setdiff(sort(IV_bayes_score, index.return = TRUE, decreasing = TRUE)$ix, c(X_idx, Y_idx))

cat("IV selection:")
cat("-------------")
for (i in IV_ranked) {
  cat(sprintf("[prob = %.2f] %s\n", IV_bayes_score[i], var_names[i]))
}

prob_hat <- rep(0, num_clusters)
for (i in seq_along(out$models)) {
  prob_hat <- prob_hat + out$models[[i]]$prob
}
prob_hat <- prob_hat / length(out$models)
cat("Mixture of Gaussians distribution:\n")
cat("----------------------------------")
for (i in seq_len(num_clusters)) {
  cat(sprintf("[%2d] %.2f\n", i, prob_hat[i]))
}

x11(); plot(density(out$ace), xlim = xl, xlab = "differential causal effect", ylab = "Posterior density", main = "IV-TETRAD, single choice", cex.main = 2.5, cex.lab = 2, cex.axis = 2)
s_ace <- sort(out$ace)
ace_tail <- 0.025
cat(sprintf("0.95 interval: [%.2f, %.2f]\n", s_ace[round(length(s_ace) * ace_tail)], s_ace[round(length(s_ace) * (1 - ace_tail))]))

equivalent_aces <- full_posterior_effects(out)
x11(); plot(density(equivalent_aces), xlim = xl, xlab = "differential causal effect", ylab = "Posterior density", main = "IV-TETRAD, equivalence class", cex.main = 2.5, cex.lab = 2, cex.axis = 2)
full_s_ace <- sort(equivalent_aces)
cat(sprintf("0.95 interval: [%.2f, %.2f]\n", s_ace[round(length(full_s_ace) * ace_tail)], s_ace[round(length(full_s_ace) * (1 - ace_tail))]))
