############################################################################################
# linear_iv_bayes_learn.R
#
# Bayesian counterpart of linear_iv using the Bayesian projections framework of
#
#    R. Silva and A. Kalaitzis,
#    "Bayesian inferencia via projections"
#    Statistics and Computing July 2015, Volume 25, Issue 4, pp 739-753
#
# Code by
#
#  - Ricardo Silva (ricardo@stats.ucl.ac.uk)
#
# Current version: 03/01/2017
# First version: 03/01/2017

#' @title Learn causal effect via instrumental variable discovery using Bayesian projections from mixture of Gaussians.
#' 
#' @description
#' 
#' The IV-TETRAD algorithm for discovering valid IVs and returning causal effects from a Bayesian analysis using
#' Silva and Kalaitzis' Bayesian projections applied to a mixture of Gaussians base model. A posterior distribution
#' over IV selections, conditioning sets, and causal effects is generated.
#' 
#' @references 
#' \url{http://link.springer.com/article/10.1007%2Fs11222-015-9557-6}
#'
#' @param dat matrix containing the data.
#' @param x index of treatment variable.
#' @param y index of outcome variable.
#' @param base_model prior usable by method \code{rnmixGibbs}, specifying parameter priors and number of mixture components.
#' @param K_win number of variables to seed the swipe search.
#' @param alpha_cut maximum ACE allowed within model.
#' @param tetrad_threshold maximum tetrad discrepancy allowed within model.
#' @param mcmc_params a list containing posterior distribution over the following quantities:
#'   \item{\code{num_steps}}{the number of MCMC iterations for Bayesian projections.}
#'   \item{\code{burn_in}}{the number of burn in iterations.}
#'   \item{\code{thinning}}}{the number of MCMC generations to skip when keeping the simulations.} 
#' @param latent_idx array of indices of variables to ignore.
#' @param store_model if \code{TRUE}, stores the MCMC samples used for Bayesian projections.
#' @param verbose verbosity level, take values 0 (complete silence), 1 or 2.
#'        
#' @return An list containing posterior distribution over the following quantities:
#'   \item{\code{ace}}{the differential causal effect of treatment on outcome.}
#'   \item{\code{IV}}{the selected instrumental variables.}
#'   \item{\code{Z}}{the selected conditional variables.}
#'
#' @export

iv_tetrad_bayes <- function(dat, x, y, K_win, alpha_cut, tetrad_threshold, base_model, mcmc_params, latent_idx = c(), store_model = FALSE, verbose = FALSE)
{
  # Base model inference
  
  if (nrow(dat) > 0) {
    if (!is.null(latent_idx)) {
      dat_obs <- dat[, -latent_idx]
    } else {
      dat_obs <- dat
    }
    samples <- rnmixGibbs(Data = list(y = dat_obs), Prior = base_model, 
                          Mcmc = list(R = mcmc_params$num_steps, keep = mcmc_params$thinning))
  } else {
    # Sample from the prior
    samples <- mog_model_sample(ncol(dat), base_model, round(mcmc_params$num_steps / mcmc_params$thinning))
  }
  
  models <- list()
  
  # General initialization
  
  thinned_burn_in <- round(mcmc_params$burn_in / mcmc_params$thinning)
  num_keep <- length(samples$nmix$compdraw) - thinned_burn_in
  ace <- rep(0, num_keep)
  IV <- list()
  Z <- list()
  score <- rep(0, num_keep)
  p <- ncol(dat)
  W_space <- setdiff(1:p, c(latent_idx, x, y))
  W <- list()
  Z <- list()
  aces <- list()
  scores <- list()
  score <- rep(0, num_keep)
  
  # Inference via IV-TETRAD projection
  
  for (i in seq_len(num_keep)) {
    if (verbose > 0) {
      cat(sprintf("[%d] IV-TETRAD projection\n", i))
    }
    model <- mog(samples, i + thinned_burn_in, latent_idx)
    out <- iv_tetrad_model(model, W_space, x, y, K_win, alpha_cut, tetrad_threshold, verbose = verbose > 1)
    ace[i] <- out$ace
    W[[i]] <- out$W
    Z[[i]] <- out$Z
    aces[[i]] <- out$aces
    score[i] <- out$score
    scores[[i]] <- out$scores
    if (store_model) {
      models[[i]] <- model
    }
  }
  
  # Return
  
  return(list(ace = ace, W = W, Z = Z, aces = aces, score = score, scores = scores, models = models))
}

# Sample from a mixture of Gaussians prior

mog_model_sample <- function(p, base_model, num_steps)
{
  K <- base_model$ncomp
  if (is.null(base_model$A)) {
    A <- 0.01
  } else {
    A <- base_model$A[1, 1]
  }
  if (is.null(base_model$nu)) {
    nu <- p + 3
  } else {
    nu <- p
  }
  if (is.null(base_model$V)) {
    V <- diag(rep(nu, p))
    inv_V <- diag(rep(1 / nu, p))
  } else {
    V <- base_model$V
    inv_V <- solve(base_model$V)
  }
  
  compdraw <- list()
  probdraw <- matrix(nrow = num_steps, ncol = K)
  for (i in seq_len(num_steps)) {
    compdraw[[i]] <- list()
    for (k in seq_len(K)) {
      inv_Sigma <- rWishart(1, nu, inv_V)[, , 1]
      inv_cov_mu <- inv_Sigma * A
      compdraw[[i]][[k]] <- list(mu = solve(chol(inv_cov_mu), rnorm(p)),
                                 rooti = t(chol(inv_Sigma)))
    }
    probdraw[i, ] <- t(rdirichlet(base_model$a))
  }
  nmix <- list(compdraw = compdraw, probdraw = probdraw)
  return(list(nmix = nmix))
}

#' @title Extract a mixture of Gaussians model out of a posterior sample.
#' 
#' @description
#' 
#' Given a posterior distribution over mixture of Gaussian parameters, generates an object
#' corresponding to one of these samples.
#' 
#' @param samples output of a call to \code{rnmixGibbs}.
#' @param id position in the sample corresponding to the object to be generated.
#' @param latent_idx position of possible latent variables for which a placeholder of zero should be added to the model.
#'        
#' @return An list containing several outcomes of interest:
#'   \item{\code{m}}{the means of each component, where each column corresponds to a different component.}
#'   \item{\code{S}}{a list of covariance matrices, where each entry corresponds to a different component.}

mog <- function(samples, id, latent_idx = c())
{
  p <- length(samples$nmix$compdraw[[1]][[1]]$mu) + length(latent_idx)
  obs_idx <- setdiff(1:p, latent_idx)
  K <- ncol(samples$nmix$probdraw)
  
  m <- matrix(rep(0, p * K), nrow = p)
  S <- list()
  for (k in 1:K) {
    m[obs_idx, k] <- samples$nmix$compdraw[[id]][[k]]$mu
    S[[k]] <- matrix(rep(0, p^2), nrow = p)
    U <- solve(samples$nmix$compdraw[[id]][[k]]$rooti)
    S[[k]][obs_idx, obs_idx] <- t(U) %*% U
  }
  prob <- matrix(samples$nmix$probdraw[id, ], ncol = 1)
  
  obj <- list(m = m, S = S, prob = prob)
  class(obj) <- "mog"
  
  return(obj)
}

#' @title Learn causal effect via instrumental variable discovery for a given mixture of Gaussians model.
#' 
#' @description
#' 
#' The IV-TETRAD algorithm for discovering valid IVs and returning causal effects from a given mixture of Gaussian
#' models. The model is given, so no statistical uncertainty is implied by this search.
#' 
#' @param model MoG object, including fields for probability of components, means and covariance matrices.
#' @param W_space index of possible instrumental variables and conditioning variables.
#' @param x index of treatment variable.
#' @param y index of outcome variable.
#' @param K_win number of variables to seed the swipe search.
#' @param alpha_cut maximum ACE allowed within model.
#' @param tetrad_threshold maximum tetrad discrepancy allowed within model.
#' @param verbose indicates whether to display messages while searching for solution.
#'        
#' @return The causal effect, selected instrumental variable set and conditioning set.
#'
#' @export

iv_tetrad_model <- function(model, W_space, x, y, K_win, alpha_cut, tetrad_threshold, verbose = FALSE)
{
  mog_moments <- moments_from_mog(model)

  # Swipe W_space to maximize tetrads. Weakly associated variables will induce high
  # alpha_hats, essentially taking care of Markov blanket selection as part of this step.
  
  num_swipes <- 0
  W <- list()
  Z <- list()
  tetrad_scores <- c()
  
  alpha_hat <- get_alpha_hat(W_space, x, y, mog_moments$corrmog)
  candidates <- setdiff(W_space, c(W_space[which(abs(alpha_hat) > alpha_cut)], W_space[which(is.nan(alpha_hat))]))
  
  while (length(candidates) >= K_win) {
    
    num_swipes <- num_swipes + 1
    
    if (verbose) {
      cat(sprintf("--------- TETRAD SEARCH, SWIPE #%d\n", num_swipes))
    }
    
    alpha_out <- alpha_search(K_win, W_space, x, y, mog_moments$corrmog, alpha_hat, candidates, TRUE)
    W[[num_swipes]] <- alpha_out$W
    Z[[num_swipes]] <- setdiff(W_space, W[[num_swipes]])
    tetrad_scores <- c(tetrad_scores, alpha_out$tetrad_score)
    candidates <- setdiff(candidates, alpha_out$W)
    
  }
  
  # Return
  
  aces <- rep(0, num_swipes)
  score_sel <- rep(0, num_swipes)
  tetrad_tests <- solution_test_model(mog_moments$corrmog, x, y, W, Z, tetrad_threshold)
  for (s in seq_len(num_swipes)) {
    aces[s] <- IV_TSLS(W[[s]], Z[[s]], x, y, mog_moments$covmog)
    score_sel[s] <- tetrad_scores[s]
  }
  score_sel[tetrad_tests$passed == FALSE] <- -Inf
  ace <- aces[which.max(score_sel)]
  if (sum(tetrad_tests$passed) == 0) {
    ace <- 0
    W <- list()
    Z <- list()
  }
  
  return(list(ace = ace, W = W, Z = Z, aces = aces, tetrad_scores = tetrad_scores, score = max(score_sel), scores = score_sel, tetrad_tests = tetrad_tests))  
}

#' @title Compute implied first and second moments of a mixture of Gaussians.
#' 
#' @description
#' 
#' Given a model parameterized as a mixture of Gaussians, compute the corresponding mean and covariance
#' matrix of the model.
#' 
#' @param model \code{mog} object, including fields for probability of components, means and covariance matrices.
#'        
#' @return The implied overall first two moments and covariance matrix of the model.

moments_from_mog <- function(model)
{
  p <- nrow(model$m)
  K <- ncol(model$m)
  first <- model$m %*% model$prob
  second <- matrix(rep(0, p^2), ncol = p)
  for (k in seq_len(K)) {
    mean_k <- model$m[, k]
    second <- second + model$prob[k] * (model$S[[k]] + mean_k %*% t(mean_k))
  }
  covmog <- second - first %*% t(first)
  sdmog <- sqrt(diag(covmog))
  corrmog <- covmog / (sdmog %*% t(sdmog)) 
  return(list(first = first, second = second, covmog = covmog, corrmog = corrmog))
}

# solution_test_model:
#
# Assess the fitness of the tetrad constraints of a particular set based on a population
# covariance matrix.
#
# Input:
#
# - model_corr: model correlation matrix
# - x, y: treatment and outcome
# - W: candidate instruments
# - Z: conditioning set
# - tetrad_threshold: maximum value of tetrad discrepancy allowed
# - verbose: if TRUE, show steps of the testing
#
# Output:
#
# - passed: TRUE, if corresponding set (W, Z) passes the implied tetrad; FALSE otherwise
# - score_tetrads: corresponding negative discrepancy

solution_test_model <- function(model_corr, x, y, W, Z, tetrad_threshold, verbose = FALSE)
{
  num_solutions <- length(W)
  passed <- rep(FALSE, num_solutions)  
  N <- nrow(dat)
  score_tetrads <- rep(0, num_solutions)
  
  for (i in seq_len(num_solutions)) {
    
    if (verbose) cat(sprintf("Testing set %d out of %d\n", i, num_solutions))
    
    tetrad_test <- tetrad_set_test_model(W[[i]], Z[[i]], x, y, model_corr)
    pass_tetrad <- mean(abs(tetrad_test) <= tetrad_threshold, na.rm = TRUE) > 0.5
    score_tetrads[i] <- -mean(abs(tetrad_test))
    passed[i] <- pass_tetrad
    
  }
  
  return(list(passed = passed, score_tetrads = score_tetrads))
}

tetrad_set_test_model <- function(W, Z, x, y, C)
{
  p <- length(W)
  if (p < 2) {
    return()
  }
  w_count <- 0
  tetrad_values <- rep(0, p * (p - 1) / 2)
  
  for (w1i in 1:(p - 1)) {
    w1 <- W[w1i]
    for (w2i in (w1i + 1):p) {
      w2 <- W[w2i]
      w_count <- w_count + 1
      tetrad_values[w_count] <- tetrad_value(C, w1, w2, x, y, Z)
    }
  }
  
  return(tetrad_values)
}

#' @title Summarize the choice of instruments picked by Bayesian \code{IV-TETRAD}.
#' 
#' @description
#' 
#' Given the output of a call to \code{iv_tetrad_bayes}, this returns an estimate of the
#' probability of picking variables as instruments.
#' 
#' @param bayes_out the output of a call to \code{iv_tetrad_bayes}
#'        
#' @return A matrix indicating where each variable was picked as an instrument.

bayes_iv_summary <- function(bayes_out)
{
  IVs <- bayes_out$W
  n <- length(IVs)
  num_v <- 0
  for (i in seq_len(n)) {
    v_choice <- which.max(bayes_out$scores[[i]])
    if (max(IVs[[i]][[v_choice]]) > num_v) {
      num_v <- max(IVs[[i]][[v_choice]])
    }
  }
  prop <- matrix(rep(0, n * num_v), nrow = n)
  for (i in seq_len(n)) {
    v_choice <- which.max(bayes_out$scores[[i]])
    prop[i, IVs[[i]][[v_choice]]] <- prop[i, IVs[[i]][[v_choice]]] + 1
  }
  return(prop)
}

#' @title Generate posterior distribution over equivalence classes of \code{IV-TETRAD}.
#' 
#' @description
#' 
#' Given the output of a call to \code{iv_tetrad_bayes}, this returns a sample of the
#' posterior distribution of effects using the whole equivalence class. A data-independent
#' prior is used to sample uniformly among the elements of the equivalence class.
#' 
#' @param bayes_out the output of a call to \code{iv_tetrad_bayes}
#'        
#' @return A vector with the given samples.

full_posterior_effects <- function(bayes_out)
{
  n <- length(bayes_out$W)
  ace_post <- rep(0, n)
  
  for (i in seq_len(n)) {
    ace_post[i] <- sample(out$aces[[i]], 1)
  }
  
  return(ace_post)
}

