############################################################################################
# synthetise.R
#
# Tools for simulation studies.
#
# Code by
#
#  - Ricardo Silva (ricardo@stats.ucl.ac.uk)
#
# Current version: 18/10/2016
# First version: 18/10/2016

# generate_sem:
#
# Generates a structural equation model from a given DAG G and other relevant parameters that
# constraint how parameters will be randomly generate. In this graph there are a treatment X
# and outcome Y variables, which are not parents of any other variables. There are two
# unmeasured confounders of X and Y, one of which is indepedent of other variables and one of
# which is associated with. The former will not shrink in variance given the other variables,
# which is a behaviour that makes inference more sensitive to biased adjustments in a different
# way from the associated confounder.
#
# Input:
#
# - err_max, err_min: maximum and minimum variance of eeach error term, to be sampled uniformly.
# - G: graph of the model as a adjacency matrix
# - X_idx, Y_idx: index, in G, of the treatment and outocme respectively.
# - U_idx: index, in G, of the X-Y latent confounder that is associated with other variables
# - U_idx2: index, in G, of the X-Y latent confounder that is independent of other variables
# - U_var: proportion of variance in X (and Y) that is due to latent variable U_idx
# - U_var2: proportion of variance in X (and Y) that is due to latent variable U_idx2
#
# Output:
#
# - L: matrix of loadings, where L[i, j] is the coefficient of parent j on child i
# - Psi: covariance matrix of error terms
# - Sigma: implied covariance matrix of the model

generate_sem <- function(err_max, err_min, G, X_idx, Y_idx, U_idx, U_idx2, U_var, U_var2)
{
  num_V <- ncol(G)
  Psi <- matrix(rep(0, num_V^2), ncol = num_V)
  L <- matrix(rep(0, num_V^2), ncol = num_V)
  Sigma <- matrix(rep(0, num_V^2), ncol = num_V)
  graph_edge <- c(rbind(col(G)[G == 1], row(G)[G == 1]))
  V_order <- as.numeric(topological.sort(graph(graph_edge)))
  V_order <- V_order[-which(V_order == X_idx)]
  V_order <- V_order[-which(V_order == Y_idx)]
  V_order <- c(V_order, X_idx, Y_idx)
  
  v_now <- c()
  
  for (v in V_order) {
    
    par <- which(G[v, ] == 1)
    if (length(par) == 0) {
      Psi[v, v] <- 1
      Sigma[v, v] <- 1
      v_now <- c(v_now, v)
      next
    }
    
    coeff <- as.matrix(rnorm(length(par)))
    
    if (v == X_idx || v == Y_idx) {
      
      which_u   <- which(par == U_idx) 
      which_u2  <- which(par == U_idx2)
      which_u12 <- c(which_u, which_u2)
      other_p   <- par[-which_u12]
      
      coeff[which_u]  <- sqrt(U_var)
      coeff[which_u2] <- sqrt(U_var2)
      
      Psi[v, v] <- (1 - U_var - U_var2) * runif(1) * (err_max - err_min) + err_min
      
      if (v == X_idx) {
        
        not_u <- setdiff(which(Sigma[par, U_idx] != 0), which_u12)
        if (length(not_u) == 1) {
          # Very special case, a single non-latent parent of X
          ax <- 1
          bx <- 2 * sqrt(U_var) * Sigma[par[not_u], U_idx]
          cx <- U_var + U_var2 + Psi[v, v] - 1
          sx <- as.numeric(2 * (runif(1) <= 0.5) - 1)
          coeff[not_u] <- (-bx + sx * sqrt(bx^2 - 4 * ax * cx)) / (2 * ax)
        } else {
          if (length(not_u) > 0) {
            target_p <- sample(not_u, 1)
            coeff[target_p] <- -sum(coeff[-c(target_p, which_u12)] * Sigma[par[-c(target_p, which_u12)], U_idx]) / 
              Sigma[par[target_p], U_idx]   
          }
          var_now <- as.numeric(t(coeff[-which_u12]) %*% Sigma[other_p, other_p] %*% coeff[-which_u12])
          coeff[-which_u12] <- sqrt(1 - Psi[v, v] - U_var - U_var2) * coeff[-which_u12] / sqrt(var_now)
          if (length(par) == 2) {
            Psi[v, v] <- 1 - U_var - U_var2
          }
        }
        
      } else if (v == Y_idx) {
        
        which_x <- which(par == X_idx)
        bx <- 2 * coeff[which_u] * Sigma[X_idx, U_idx] + 2 * coeff[which_u2] * Sigma[X_idx, U_idx2]
        cx <- U_var + U_var2 + Psi[v, v] - 1
        delta <- bx^2 - 4 * cx
        max_coeff_x <- 0.5 * (-bx + sqrt(delta))
        min_coeff_x <- 0.5 * (-bx - sqrt(delta))
        
        if (max_coeff_x - min_coeff_x < 0.05) {
          coeff[which_x] <- runif(1) * (max_coeff_x - min_coeff_x) + min_coeff_x
        } else {
          while (coeff[which_x] > max_coeff_x || coeff[which_x] < min_coeff_x) {
            coeff[which_x] <- rnorm(1)
          }
        }
        
        which_u12x <- c(which_u12, which_x)
        not_u12x <- setdiff(which(Sigma[par, X_idx] != 0), which_u12x)
        if (length(not_u12x) == 1) {
          stop("Graph has structure we cannot simulate from: outcome has only one other parent besides X, U1 and U2")  
        }
        if (length(not_u12x) > 0) {
          target_p <- sample(not_u12x, 1)
          other_p <- par[-which_u12x]
          #coeff[target_p] <- -sum(coeff[-c(target_p, which_u12x)] * Sigma[par[-c(target_p, which_u12x)], X_idx]) / 
          #                     Sigma[par[target_p], X_idx] 
          coeff[target_p] <- -(sum(coeff[-c(target_p, which_u12x)] * coeff[which_x] * Sigma[par[-c(target_p, which_u12x)], X_idx]) +
                                 sum(coeff[-c(target_p, which_u12x)] * coeff[which_u] * Sigma[par[-c(target_p, which_u12x)], U_idx])) /
            (coeff[which_x]  * Sigma[par[target_p], X_idx] + coeff[which_u] * Sigma[par[target_p], U_idx])
          var_now <- as.numeric(t(coeff[-which_u12x]) %*% Sigma[other_p, other_p] %*% coeff[-which_u12x])
          coeff[-which_u12x] <- sqrt(1 - Psi[v, v] - coeff[which_x]^2 - U_var - U_var2 - 2 * coeff[which_x] * coeff[which_u] * Sigma[X_idx, U_idx] - 2 * coeff[which_x] * coeff[which_u2] * Sigma[X_idx, U_idx2]) * 
            coeff[-which_u12x] / sqrt(var_now)
        } else {
          # Only three parents, X, U1 and U2. Scale X ->Y coefficient or Psi             
          if (1 - as.numeric(t(coeff) %*% Sigma[par, par] %*% coeff) < 0) {
            bv <- 2 * (coeff[which_u] * Sigma[X_idx, U_idx] + coeff[which_u2] * Sigma[X_idx, U_idx2])
            cv <- 1 - U_var - U_var2 - Psi[v, v]
            delta <- bv^2 - 4 * cv
            if (delta >= 0) {
              if (runif(1) < 0.5) {
                coeff[1] <- (-bv + sqrt(delta)) / 2
              } else {
                coeff[1] <- (-bv - sqrt(delta)) / 2
              }
            }
          }
          Psi[v, v] <- 1 - as.numeric(t(coeff) %*% Sigma[par, par] %*% coeff)
        }
        
      } else {
        
        stop("Graph has structure we cannot simulate from")
        
      }
      
    } else { 
      
      Psi[v, v] <- runif(1) * (err_max - err_min) + err_min  
      impl_var <- as.numeric(t(coeff) %*% Sigma[par, par] %*% coeff)
      coeff <- sqrt(1 - Psi[v, v]) * coeff / sqrt(impl_var)
      
    }
    
    L[v, par] <- coeff
    Sigma[v, v_now] <- L[v, v_now] %*% Sigma[v_now, v_now]
    Sigma[v_now, v] <- Sigma[v, v_now]
    Sigma[v, v] <- L[v, v_now] %*% Sigma[v_now, v_now] %*% t(L[v, v_now, drop = FALSE]) + Psi[v, v]
    v_now <- c(v_now, v)
    
  }
  
  return(list(L = L, Psi = Psi, Sigma = Sigma))
}

# generate_data:
#
# Given an existing SEM, generates the corresponding data. As the input SEM only specifies
# second moments, further input is necessary to indicate what the distribution of the 
# error terms should be.
#
# Input:
#
# - N: sample size
# - problem: the output of a problem generating function such as simulate_partition_problem
# - laplace_error: if TRUE, generates error terms according to a Laplace distribution of
#                  variance estipulated by the input SEM. If FALSE, increase non-Gaussianity
#                  by taking a Laplace random variable to the power of 1.5. THE IMPLIED VARIANCE
#                  is only approximated here.

generate_data <- function(N, problem, laplace_error) 
{
  G <- problem$G
  model <- problem$model
  V_order <- as.numeric(topological.sort(graph(c(rbind(col(G)[G == 1], row(G)[G == 1])))))
  
  num_V <- ncol(G)
  D <- matrix(rep(0, N * num_V), ncol = num_V)
  for (v in V_order) {
    par <- which(G[v, ] == 1)
    if (length(par) > 0) {
      D[, v] <- D[, par] %*% t(model$L[v, par, drop = FALSE])
    }
    e <- rexp(N) / sqrt(2) * (2 * (runif(N) > 0.5) - 1)
    if (laplace_error) {
      D[, v] <- D[, v] + e * sqrt(model$Psi[v, v])
    } else {
      error_terms <- abs(e)^1.5 * sign(e)
      D[, v] <- D[, v] + error_terms * sqrt(model$Psi[v, v] / 2.122)
    }
  }
  
  return(D)
}

# simulate_partition_problem:
#
# This function builds a particular structure that is useful to test causal effect estimation
# methods that exploit the discovery of instrumental variables. The structure of the generated
# networks follows this template.
#
# There are five different groups of variables:
#
# - X and Y: the treatment and outcome variables, respectively. They have two latent parents
#   in common, U_XY and U_XY2;
# - the "best" variables: these variables have a common latent parent U_0 and X as the common
#   child. U_0 creates a back-door with X and Y, but it can be blocked by conditioning on
#   everybody else;
# - the "spurious" variables: they come in K groups. Each group has as a child a Ui latent variable,
#   which is a common parent of X and Y. U_0 is also a common latent parent. These variables
#   cannot be IVs, but they are part of conditional tetrad constraints because of Ui;
# - the "direct" variables: these have X and Y as children, and U_0 as a parent. They are not 
#   IVs, but do not participate on conditional tetrad constraints;
# - the "possible" and "collider" variables: the "possible" variables have X and the "collider"
#   variables as children and U_0 as a parent. The "collider" have X as children, and the also U_XY 
#   as a common parent. So the "collider" variables cannot be IVs, and moreover conditioning on them
#   makes the "possible" variables invalid. The "possible" variables are potentially valid,
#   but an algorithm that does not search for the conditioning variables (like IV-TETRAD)
#   will not be possible;
#
# Input:
#
# - C:             number of variables per cluster
# - K:             number of clusters
# - num_direct:    number of elements in cluster that are parents of Y
# - num_possible:  number of elements in possible solution, but only if we do not condition on colliders
# - num_colliders: number of colliders destroying the tetrads of the "possible set"
# - num_best:      number of elements in "true cluster"
# - U_var:         proportion of variance in treatment and outcome due to their latent common causes possible connected to colliders
# - U_var2:        proportion of variance in treatment and outcome due to their latent common causes not connected to colliders.
# - direct_X:      if TRUE, add directed edge between "possible" variables and treatment X
#
# Output:
#
# This returns a list where the two most important components are the generated graph and
# the corresponding structural equation model. Indices for the relevant variables in the
# model (which index in G corresponds to treatment X etc.) are also important.

simulate_partition_problem <- function(C, K, num_direct, num_possible, num_colliders, num_best, U_var, U_var2 = 0, direct_X = TRUE)
{
  if (C == 0) {
    K <- 0
  }
  
  num_spurious    <- C * K
  W_spurious_idx  <- seq_len(num_spurious)
  Ui_idx          <- num_spurious + seq_len(K)
  U0_idx          <- num_spurious + K + 1
  X_idx           <- U0_idx + 1
  Y_idx           <- X_idx + 1
  UXY_idx         <- Y_idx + 1
  UXY2_idx        <- UXY_idx + 1
  W_direct_idx    <- UXY2_idx + seq_len(num_direct)
  W_possible_idx  <- UXY2_idx + num_direct + seq_len(num_possible)
  W_colliders_idx <- UXY2_idx + num_direct + num_possible + seq_len(num_colliders)
  W_best_idx      <- UXY2_idx + num_direct + num_possible + num_colliders + seq_len(num_best)
  num_V           <- UXY2_idx + num_direct + num_possible + num_colliders + num_best
  
  G <- matrix(rep(0, num_V^2), ncol = num_V)
  for (i in seq_len(K)) {
    G[num_spurious + i, (i - 1) * C + seq_len(C)] <- 1
  }
  G[c(W_spurious_idx, W_direct_idx, W_possible_idx, W_best_idx), U0_idx] <- 1
  G[W_colliders_idx,  c(W_possible_idx, UXY_idx)] <- 1
  G[X_idx,            c(Ui_idx, UXY_idx, UXY2_idx, W_direct_idx, W_possible_idx, W_colliders_idx, W_best_idx)] <- 1
  G[Y_idx,            c(X_idx, Ui_idx, UXY_idx, UXY2_idx, W_direct_idx)] <- 1
  
  if (U_var2 == 0) {
    G[X_idx, UX2_idx] <- 0
    G[Y_idx, UX2_idx] <- 0
  }
  if (!direct_X) {
    G[X_idx, W_possible_idx] <- FALSE
  }
  
  model <- generate_sem(0.3, 0, G, X_idx, Y_idx, UXY_idx, UXY2_idx, U_var, U_var2)

  return(list(G = G, model = model, 
              W_spurious_idx = W_spurious_idx, W_direct_idx = W_direct_idx, 
              W_possible_idx = W_possible_idx, W_colliders_idx = W_colliders_idx, W_best_idx = W_best_idx,
              Ui_idx = Ui_idx, U0_idx = U0_idx, UXY_idx = UXY_idx, UXY2_idx = UXY2_idx,  
              X_idx = X_idx, Y_idx = Y_idx))
}

# simulate_confounded_unblockable:
#
# The structure of this problem is:
#
# U_i exogenous
# W    <- U_i
# U_XY <- U_i
# X    <- U_XY
# Y    <- U_XY
#
# where X is the treatment, Y the outcome and W the candidate instruments. This structure means 
# that we should detect that set W is not a set of valid instruments.
#
# Input:
#
# - num_W: size of set W
# - U_var: variance of U_XY
# - U_var2: variance of other confounders of X and Y.
#
# Output:
#
# This returns a list where the two most important components are the generated graph and
# the corresponding structural equation model. Indices for the relevant variables in the
# model (which index in G corresponds to treatment X etc.) are also important.

simulate_confounded_problem_unblockable <- function(num_W, U_var, U_var2)
{
  W        <- seq_len(num_W)
  Ui_idx   <- num_W + 1
  X_idx    <- Ui_idx + 1
  Y_idx    <- X_idx + 1
  UXY_idx  <- Y_idx + 1
  UXY2_idx <- UXY_idx + 1

  num_V   <- UXY2_idx
  
  G <- matrix(rep(0, num_V^2), ncol = num_V)
  
  G[W,        Ui_idx] <- 1
  G[UXY_idx,  Ui_idx] <- 1
  G[X_idx,    c(UXY_idx, UXY2_idx)] <- 1
  G[Y_idx,    c(X_idx, UXY_idx, UXY2_idx)] <- 1
  
  model <- generate_sem(0.3, 0.1, G, X_idx, Y_idx, UXY_idx, UXY2_idx, U_var, U_var2)
  
  return(list(G = G, model = model, W_possible_idx = W, Ui_idx = Ui_idx, UXY_idx = UXY_idx, UXY2_idx = UXY2_idx,  
              X_idx = X_idx, Y_idx = Y_idx))
}

# simulate_confounded_problem_blockable:
#
# The structure of this problem is:
#
# U_i exogenous
# W    <- U_i
# X    <- U_XY
# Y    <- U_XY
#
# where X is the treatment, Y the outcome and W the candidate instruments. This structure means 
# that we should detect that set W is a set of valid instruments. But this requires that for each
# w in W, we need to condition on W \{w}.
#
# Input:
#
# - num_W: size of set W
# - U_var: variance of U_XY
# - U_var2: variance of other confounders of X and Y.
#
# Output:
#
# This returns a list where the two most important components are the generated graph and
# the corresponding structural equation model. Indices for the relevant variables in the
# model (which index in G corresponds to treatment X etc.) are also important.

simulate_confounded_problem_blockable <- function(num_W, U_var, U_var2)
{
  W        <- seq_len(num_W)
  Ui_idx   <- num_W + 1
  X_idx    <- Ui_idx + 1
  Y_idx    <- X_idx + 1
  UXY_idx  <- Y_idx + 1
  UXY2_idx <- UXY_idx + 1
  
  num_V   <- UXY2_idx
  
  G <- matrix(rep(0, num_V^2), ncol = num_V)
  
  G[W,        Ui_idx] <- 1
  G[X_idx,    c(W, UXY_idx, UXY2_idx)] <- 1
  G[Y_idx,    c(X_idx, UXY_idx, UXY2_idx)] <- 1
  
  model <- generate_sem(0.3, 0, G, X_idx, Y_idx, UXY_idx, UXY2_idx, U_var, U_var2)
  
  return(list(G = G, model = model, W_possible_idx = W, UXY_idx = UXY_idx, UXY2_idx = UXY2_idx,  
              X_idx = X_idx, Y_idx = Y_idx))
}


